package controller.model;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.BlockingQueue;

import controller.DriveModul;
import controller.PrototypeFirstTime;

/**
 * Trieda vytvarajuca zaznamy o zakrutach. Bezi pri tom v inom vlakne
 * ako ovladaci modul vozidla.
 * 
 * @author Ivan Valencik
 */
public class TurnDetector implements Runnable
{
	private final BlockingQueue<TickInfo> tickInfoPipe;
	
	private static final LinkedList<Turn> turns = new LinkedList<Turn>();

	// odporucam tento atribut nepouzivat na rozoznavanie, pouzivany je na synchronizaciu
	// a jeho vyznam sa pri zmene jej implementacie moze zmenit
	private static double _focusValue;

	public TurnDetector(BlockingQueue<TickInfo> source)
	{
		tickInfoPipe = source;
	}

	public void run()
	{
		try {
			while (true)
			{
				TickInfo tick = tickInfoPipe.take();
				
				inform(tick);
				//System.out.println("Dlzka radu na spracovanie: " + tickInfoPipe.size());
			}
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri prijmani akcie do modelu trate.");
			e.printStackTrace();
		}
	}
		
	/**
	 * Metoda vrati zoznam rozoznanych zakrut. <br/><br/> 
	 * Pozor: zoznam sa sa neaktualizuje, pre ziskanie jeho novej verzie je
	 * metodu potrebne zavolat znova.
	 * @return Zoznam zakrut
	 */
	@SuppressWarnings("unchecked")
	public static LinkedList<Turn> getTurnList(double trackWidth, double friction)
	{
		LinkedList<Turn> tmp;
		synchronized (turns)
		{
			tmp =(LinkedList<Turn>) turns.clone();
		}
		for (Turn turn : tmp) {
			turn.setEnterPosition(trackWidth, friction);
			if(turn.getPeakValue()>(trackWidth>14 ? 0.18:0.1))
				turn.setBreakDist(DriveModul.countDistanceFromSpeed(turn.getSpeed(), friction));
			else
				turn.setBreakDist(200);
		}
		return tmp;
	}

	/**
	 * <p>Metoda vrati hodnotu focusu na urceny uhol. Metoda je blokujuca.</p>
	 * <p>Kazde volanie metody vrati hodnotu ASAP, avsak meranie moze byt vykonane
	 * iba kazdu sekundu herneho casu ( [senzor_model].getCurrentLapTime() ).
	 * @param desiredFocus zelane nastavenie focusu z rozsahu <-90, 90>
	 * @throws TooSoonException vynimka je vyhodena ak bola metoda zavolana skorej ako za
	 * jednu sekundu herneho casu
	 */
	private double requestSpecificFocusValue(int desiredFocus) throws TooSoonException
	{
		PrototypeFirstTime.requestFocusChange(desiredFocus);

		synchronized(PrototypeFirstTime.lock)
		{
			try {
				PrototypeFirstTime.lock.wait();
				if (_focusValue == -3)
					throw new TooSoonException();
				else
					return _focusValue;
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	/**
	 * Tato metoda ma byt volana len v ramci ziadania nastavenia noveho focusu.
	 * @param focusValue hodnota stredneho senzora z focusu
	 */
	public static void setFocusValue(double focusValue)
	{
		_focusValue = focusValue;
	}
	
	private LinkedList<Double> steeringRecords = null;
	private LinkedList<Double> speedRecords = null;
	private double turnStart;
	private double speed;
	
	private static final double MAX_DIFF_RANGE = 0.01;
	private static final int SEQUENCE_LENGTH = 5;
		
	/**
	 * Metoda spravi priemer honot postupnosti, ale iba pokial su vsetky prvky
	 * v rozsahu urcenom konstantou MAX_DIFF_RANGE.
	 * @param subList zoznam hodnot zatoceni volantom
	 * @return priemerna hodnota postupnosti alebo -1, ak nesplnala podmienky
	 */
	private double checkValues(List<Double> subList) {
		double min = 1, max = 0;

		for (double d : subList) {
			if ((d > 0 && subList.get(0) < 0) || (d < 0 && subList.get(0) > 0))
				return -1;

			double d1 = Math.abs(d);
			if (d1 < min)
				min = d1;
			if (d1 > max)
				max = d1;
		}
		
		if (max - min < MAX_DIFF_RANGE) {
			double sum = 0;
			for (double d : subList)
				sum += d;
			return sum / subList.size();
		}

		return -1;
	}
	
	/**
	 * Vypocita priemernu rychlost pri v sledovanom case.
	 * @param startIndex Tik zakruty od ktoreho ma byt rychlost vypocitana.
	 * @return Priemer hodnot rychlost v useku najprudsieho zatacania.
	 */
	private double computeAvgSpeed(int startIndex) {
		double sum = 0;
		
		if (speedRecords.size() < SEQUENCE_LENGTH) {
			for (Double d : speedRecords)
				sum += d;
			return sum / speedRecords.size();
		}
		
		for (int i = startIndex; i < startIndex + SEQUENCE_LENGTH; i++)
			sum += speedRecords.get(i);
		
		return sum / SEQUENCE_LENGTH;
	}

	/**
	 * Na zaklade hodnot natoceni volantu v zakrute priradi k zakrute cislo reprezentujuce
	 * usek najprudsieho zatacania v zakrute.
	 * @return hodnota reprezentujuca zatocenie zakruty
	 */
	private double processSteeringRecord() {
		double val = 0;
		double maxVal = 0;

		for (int i = 0; i <= steeringRecords.size() - SEQUENCE_LENGTH; i++)
			if ((val = checkValues(steeringRecords.subList(i, i + SEQUENCE_LENGTH - 1))) != -1)
			{				
				if (Math.abs(val) > Math.abs(maxVal)) {
					maxVal = val;
					speed = computeAvgSpeed(i);
				}
			}
		
		return maxVal;
	}

	/**
	 * Metoda rozoznavajuca zakruty na zaklade riadenim vykonanych cinnosti.
	 * @param tick Informacie o tiku
	 * @throws InterruptedException vynimka vzniknuta pri cakani na zmenu focusu
	 */
	public void inform(TickInfo tick)
	{
		if (steeringRecords == null && tick.isTurning()) {
			//System.out.println("zaciatok zakruty");
			steeringRecords = new LinkedList<Double>();
			speedRecords = new LinkedList<Double>();
			steeringRecords.add(tick.getAction().steering);
			speedRecords.add(tick.getSensorModel().getSpeed());
			turnStart = tick.getSensorModel().getDistanceFromStartLine();
		}

		if (steeringRecords != null && tick.isTurning()) {
			steeringRecords.add(tick.getAction().steering);
			speedRecords.add(tick.getSensorModel().getSpeed());
		}

		if (steeringRecords != null && !tick.isTurning()) {
			//System.out.println("koniec zakruty");
			double val = processSteeringRecord();
			
			// vypis pre pracu v MatLAB
//			System.out.println("ZACIATOK ZAKRUTY");
//			for (Double d : steeringRecord)
//				System.out.print(d + ", ");
//			System.out.println();
//			System.out.println("KONIEC ZAKRUTY");
			
			steeringRecords = null;
			speedRecords = null;
			
			// zaznamenanie dlzky roviny po zakrute
			double angle = tick.getSensorModel().getAngleToTrackAxis();
			double frontEdgeDist = -1;
			try {
				frontEdgeDist = requestSpecificFocusValue((int)(angle * 90));
			} catch (TooSoonException e) {
				//System.out.println("priskoro");
			}
			
			// pridanie zaznamu o zakrute
			synchronized (turns)
			{
				turns.add(new Turn(turnStart,
						tick.getSensorModel().getDistanceFromStartLine(), val,
						speed, frontEdgeDist));
				
				System.out.println(turns.getLast().getDirection().toString() + ": " +
						turns.getLast().getPeakValue() + "\n peak spd: " +
						turns.getLast().getPeakSpeed() + " edge dist:" +
						turns.getLast().getFrontEdgeDistance() + "\n");
			}
		}
		
	}
}

/**
 * Vynimka upozornujca na to, ze metoda, ktora ju vyhodila bola zavolana skorej,
 * ako je dovolene.
 * @author Ivan Valencik
 */
@SuppressWarnings("serial")
class TooSoonException extends Exception {}

